@@ -1,25 +1,20 @@
 import base64
-import datetime
 import json
 import logging
 import re
-import time
-from typing import Any, Dict, Optional, Tuple, Union
+from typing import Any, Dict, Union
 from urllib.parse import urljoin
 
-import pytz
 import requests
 from flask import Response as FlaskResponse
 from jsonschema import ValidationError, validate
-from moto.apigateway.models import apigateway_backends
 from requests.models import Response
 
 from localstack import config
 from localstack.constants import (
     APPLICATION_JSON,
+    HEADER_LOCALSTACK_AUTHORIZATION,
     HEADER_LOCALSTACK_EDGE_URL,
-    LOCALHOST_HOSTNAME,
-    PATH_USER_REQUEST,
     TEST_AWS_ACCOUNT_ID,
 )
 from localstack.services.apigateway import helpers
@@ -32,13 +27,11 @@ from localstack.services.apigateway.helpers import (
     PATH_REGEX_PATH_MAPPINGS,
     PATH_REGEX_RESPONSES,
     PATH_REGEX_TEST_INVOKE_API,
+    PATH_REGEX_USER_REQUEST,
     PATH_REGEX_VALIDATORS,
-    apply_integration_response_template,
-    apply_template,
     extract_path_params,
     extract_query_string_params,
     get_cors_response,
-    get_resource_for_path,
     handle_accounts,
     handle_authorizers,
     handle_base_path_mappings,
@@ -49,7 +42,12 @@ from localstack.services.apigateway.helpers import (
     handle_vpc_links,
     make_error_response,
 )
-from localstack.services.apigateway.integration import SnsIntegration
+from localstack.services.apigateway.integration import (
+    RequestTemplates,
+    ResponseTemplates,
+    SnsIntegration,
+    VtlTemplate,
+)
 from localstack.services.awslambda import lambda_api
 from localstack.services.generic_proxy import ProxyListener
 from localstack.services.kinesis import kinesis_listener
@@ -63,8 +61,7 @@ from localstack.utils.aws.aws_responses import (
     request_response_stream,
     requests_response,
 )
-from localstack.utils.aws.request_context import MARKER_APIGW_REQUEST_REGION, THREAD_LOCAL
-from localstack.utils.common import camel_to_snake_case, json_safe, long_uid, to_bytes, to_str
+from localstack.utils.common import camel_to_snake_case, json_safe, to_bytes, to_str
 
 # set up logger
 from localstack.utils.http import add_query_params_to_url
@@ -76,17 +73,6 @@ TARGET_REGEX_PATH_S3_URI = (
     r"^arn:aws:apigateway:[a-zA-Z0-9\-]+:s3:path/(?P<bucket>[^/]+)/(?P<object>.+)$"
 )
 TARGET_REGEX_ACTION_S3_URI = r"^arn:aws:apigateway:[a-zA-Z0-9\-]+:s3:action/(?:GetObject&Bucket\=(?P<bucket>[^&]+)&Key\=(?P<object>.+))$"
-# regex path pattern for user requests
-PATH_REGEX_USER_REQUEST = (
-    r"^/restapis/([A-Za-z0-9_\-]+)(?:/([A-Za-z0-9_\-]+))?/%s/(.*)$" % PATH_USER_REQUEST
-)
-# URL pattern for invocations
-HOST_REGEX_EXECUTE_API = (
-    r"(?:.*://)?([a-zA-Z0-9-]+)\.execute-api\.(%s|([^\.]+)\.amazonaws\.com)(.*)"
-    % LOCALHOST_HOSTNAME
-)
-
-REQUEST_TIME_DATE_FORMAT = "%d/%b/%Y:%H:%M:%S %z"
 
 
 class AuthorizationError(Exception):
@@ -120,7 +106,7 @@ class ProxyListenerApiGateway(ProxyListener):
         if re.match(PATH_REGEX_PATH_MAPPINGS, path):
             return handle_base_path_mappings(method, path, data, headers)
 
-        if is_test_invoke_method(method, path):
+        if helpers.is_test_invoke_method(method, path):
             # if call is from test_invoke_api then use http_method to find the integration,
             #   as test_invoke_api makes a POST call to request the test invocation
             match = re.match(PATH_REGEX_TEST_INVOKE_API, path)
@@ -150,13 +136,15 @@ class ProxyListenerApiGateway(ProxyListener):
 
         # add missing implementations
         if response.status_code == 404:
-            data = data and json.loads(to_str(data))
             result = None
             if path == "/account":
+                data = data and json.loads(to_str(data))
                 result = handle_accounts(method, path, data, headers)
             elif path.startswith("/vpclinks"):
+                data = data and json.loads(to_str(data))
                 result = handle_vpc_links(method, path, data, headers)
             elif re.match(PATH_REGEX_CLIENT_CERTS, path):
+                data = data and json.loads(to_str(data))
                 result = handle_client_certificates(method, path, data, headers)
 
             if result is not None:
@@ -362,77 +350,8 @@ def apply_response_parameters(invocation_context: ApiInvocationContext):
     return response
 
 
-def set_api_id_stage_invocation_path(
-    invocation_context: ApiInvocationContext,
-) -> ApiInvocationContext:
-    # skip if all details are already available
-    values = (
-        invocation_context.api_id,
-        invocation_context.stage,
-        invocation_context.path_with_query_string,
-    )
-    if all(values):
-        return invocation_context
-
-    # skip if this is a websocket request
-    if invocation_context.is_websocket_request():
-        return invocation_context
-
-    path = invocation_context.path
-    headers = invocation_context.headers
-
-    path_match = re.search(PATH_REGEX_USER_REQUEST, path)
-    host_header = headers.get(HEADER_LOCALSTACK_EDGE_URL, "") or headers.get("Host") or ""
-    host_match = re.search(HOST_REGEX_EXECUTE_API, host_header)
-    test_invoke_match = re.search(PATH_REGEX_TEST_INVOKE_API, path)
-    if path_match:
-        api_id = path_match.group(1)
-        stage = path_match.group(2)
-        relative_path_w_query_params = "/%s" % path_match.group(3)
-    elif host_match:
-        api_id = extract_api_id_from_hostname_in_url(host_header)
-        stage = path.strip("/").split("/")[0]
-        relative_path_w_query_params = "/%s" % path.lstrip("/").partition("/")[2]
-    elif test_invoke_match:
-        # special case: fetch the resource details for TestInvokeApi invocations
-        stage = None
-        region_name = invocation_context.region_name
-        api_id = test_invoke_match.group(1)
-        resource_id = test_invoke_match.group(2)
-        query_string = test_invoke_match.group(4) or ""
-        apigateway = aws_stack.connect_to_service(
-            service_name="apigateway", region_name=region_name
-        )
-        resource = apigateway.get_resource(restApiId=api_id, resourceId=resource_id)
-        resource_path = resource.get("path")
-        relative_path_w_query_params = f"{resource_path}{query_string}"
-    else:
-        raise Exception(
-            f"Unable to extract API Gateway details from request: {path} {dict(headers)}"
-        )
-    if api_id:
-        # set current region in request thread local, to ensure aws_stack.get_region() works properly
-        if getattr(THREAD_LOCAL, "request_context", None) is not None:
-            THREAD_LOCAL.request_context.headers[MARKER_APIGW_REQUEST_REGION] = API_REGIONS.get(
-                api_id, ""
-            )
-
-    # set details in invocation context
-    invocation_context.api_id = api_id
-    invocation_context.stage = stage
-    invocation_context.path_with_query_string = relative_path_w_query_params
-    return invocation_context
-
-
-def extract_api_id_from_hostname_in_url(hostname: str) -> str:
-    """Extract API ID 'id123' from URLs like https://id123.execute-api.localhost.localstack.cloud:4566"""
-    match = re.match(HOST_REGEX_EXECUTE_API, hostname)
-    api_id = match.group(1)
-    return api_id
-
-
 def invoke_rest_api_from_request(invocation_context: ApiInvocationContext):
-    set_api_id_stage_invocation_path(invocation_context)
+    helpers.set_api_id_stage_invocation_path(invocation_context)
     try:
         return invoke_rest_api(invocation_context)
     except AuthorizationError as e:
@@ -449,7 +368,7 @@ def invoke_rest_api(invocation_context: ApiInvocationContext):
     # run gateway authorizers for this request
     authorize_invocation(invocation_context)
 
-    extracted_path, resource = get_target_resource_details(invocation_context)
+    extracted_path, resource = helpers.get_target_resource_details(invocation_context)
     if not resource:
         return make_error_response("Unable to find path %s" % invocation_context.path, 404)
 
@@ -497,6 +416,8 @@ def invoke_rest_api(invocation_context: ApiInvocationContext):
 def invoke_rest_api_integration(invocation_context: ApiInvocationContext):
     try:
         response = invoke_rest_api_integration_backend(invocation_context)
+        # TODO remove this setter once all the integrations are migrated to the new response
+        #  handling
         invocation_context.response = response
         response = apply_response_parameters(invocation_context)
         return response
@@ -512,7 +433,6 @@ def invoke_rest_api_integration_backend(invocation_context: ApiInvocationContext
     # define local aliases from invocation context
     invocation_path = invocation_context.path_with_query_string
     method = invocation_context.method
-    path = invocation_context.path
     data = invocation_context.data
     headers = invocation_context.headers
     api_id = invocation_context.api_id
@@ -526,6 +446,11 @@ def invoke_rest_api_integration_backend(invocation_context: ApiInvocationContext
     integration_type_orig = integration.get("type") or integration.get("integrationType") or ""
     integration_type = integration_type_orig.upper()
     uri = integration.get("uri") or integration.get("integrationUri") or ""
+    # XXX we need replace the internal Authorization header with an Authorization header set from
+    # the customer, even if it's empty that's what's expected in the integration.
+    custom_auth_header = invocation_context.headers.pop(HEADER_LOCALSTACK_AUTHORIZATION, "")
+    invocation_context.headers["Authorization"] = custom_auth_header
+
     try:
         path_params = extract_path_params(path=relative_path, extracted_path=resource_path)
         invocation_context.path_params = path_params
@@ -542,50 +467,32 @@ def invoke_rest_api_integration_backend(invocation_context: ApiInvocationContext
                     uri.split(":lambda:path")[1].split("functions/")[1].split("/invocations")[0]
                 )
 
-            # apply custom request template
-            data_str = data
-            is_base64_encoded = False
-            try:
-                data_str = json.dumps(data) if isinstance(data, (dict, list)) else to_str(data)
-                data_str = apply_template(
-                    integration,
-                    "request",
-                    data_str,
-                    path_params=path_params,
-                    query_params=query_string_params,
-                    headers=headers,
-                )
-            except UnicodeDecodeError:
-                data_str = base64.b64encode(data_str)
-                is_base64_encoded = True
-            except Exception as e:
-                LOG.warning("Unable to convert API Gateway payload to str: %s", (e))
-                pass
+            invocation_context.context = helpers.get_event_request_context(invocation_context)
+            invocation_context.stage_variables = helpers.get_stage_variables(invocation_context)
+            if invocation_context.authorizer_type:
+                authorizer_context = {
+                    invocation_context.authorizer_type: invocation_context.auth_context
+                }
+                invocation_context.context["authorizer"] = authorizer_context
+
+            request_templates = RequestTemplates()
+            payload = request_templates.render(invocation_context)
 
-            # Sample request context:
-            # https://docs.aws.amazon.com/apigateway/latest/developerguide/api-gateway-create-api-as-simple-proxy-for-lambda.html#api-gateway-create-api-as-simple-proxy-for-lambda-test
-            request_context = get_event_request_context(invocation_context)
-            stage_variables = (
-                get_stage_variables(api_id, stage)
-                if not is_test_invoke_method(method, path)
-                else None
-            )
             # TODO: change this signature to InvocationContext as well!
             result = lambda_api.process_apigateway_invocation(
                 func_arn,
                 relative_path,
-                data_str,
+                payload,
                 stage,
                 api_id,
                 headers,
-                is_base64_encoded=is_base64_encoded,
+                is_base64_encoded=invocation_context.is_data_base64_encoded,
                 path_params=path_params,
                 query_string_params=query_string_params,
                 method=method,
                 resource_path=resource_path,
-                request_context=request_context,
-                event_context=invocation_context.context,
-                stage_variables=stage_variables,
+                request_context=invocation_context.context,
+                stage_variables=invocation_context.stage_variables,
             )
 
             if isinstance(result, FlaskResponse):
@@ -619,14 +526,15 @@ def invoke_rest_api_integration_backend(invocation_context: ApiInvocationContext
                 response.multi_value_headers = parsed_result.get("multiValueHeaders") or {}
 
             # apply custom response template
-            response._content = apply_template(integration, "response", response._content)
-            response.headers["Content-Length"] = str(len(response.content or ""))
+            invocation_context.response = response
 
-            return response
+            response_templates = ResponseTemplates()
+            response_templates.render(invocation_context)
+            invocation_context.response.headers["Content-Length"] = str(len(response.content or ""))
+            return invocation_context.response
 
         raise Exception(
-            'API Gateway integration type "%s", action "%s", method "%s" invalid or not yet implemented'
-            % (integration_type, uri, method)
+            f'API Gateway integration type "{integration_type}", action "{uri}", method "{method}"'
         )
 
     elif integration_type == "AWS":
@@ -644,15 +552,11 @@ def invoke_rest_api_integration_backend(invocation_context: ApiInvocationContext
                 target = ""
 
             try:
-                data = json.dumps(data) if isinstance(data, (dict, list)) else to_str(data)
-                payload = apply_template(
-                    integration,
-                    "request",
-                    data,
-                    path_params=path_params,
-                    query_params=query_string_params,
-                    headers=headers,
-                )
+                invocation_context.context = helpers.get_event_request_context(invocation_context)
+                invocation_context.stage_variables = helpers.get_stage_variables(invocation_context)
+                request_templates = RequestTemplates()
+                payload = request_templates.render(invocation_context)
+
             except Exception as e:
                 LOG.warning("Unable to convert API Gateway payload to str", e)
                 raise
@@ -668,22 +572,20 @@ def invoke_rest_api_integration_backend(invocation_context: ApiInvocationContext
             )
 
             # apply response template
-            result = apply_request_response_templates(
-                result, response_templates, content_type=APPLICATION_JSON
-            )
-            return result
+            invocation_context.response = result
+            response_templates = ResponseTemplates()
+            response_templates.render(invocation_context)
+            return invocation_context.response
 
         elif "states:action/" in uri:
             action = uri.split("/")[-1]
 
             if APPLICATION_JSON in integration.get("requestTemplates", {}):
-                payload = apply_request_response_templates(
-                    data,
-                    integration.get("requestTemplates"),
-                    content_type=APPLICATION_JSON,
-                    as_json=True,
-                )
+                request_templates = RequestTemplates()
+                payload = request_templates.render(invocation_context)
+                payload = json.loads(payload)
             else:
+                # XXX decoding in py3 sounds wrong, this actually might break
                 payload = json.loads(data.decode("utf-8"))
             client = aws_stack.connect_to_service("stepfunctions")
 
@@ -723,9 +625,12 @@ def invoke_rest_api_integration_backend(invocation_context: ApiInvocationContext
                 response = requests_response(content=result)
 
             # apply response templates
-            response = apply_request_response_templates(
-                response, response_templates, content_type=APPLICATION_JSON
-            )
+            invocation_context.response = response
+            response_templates = ResponseTemplates()
+            response_templates.render(invocation_context)
+            # response = apply_request_response_templates(
+            #     response, response_templates, content_type=APPLICATION_JSON
+            # )
             return response
         # https://docs.aws.amazon.com/apigateway/api-reference/resource/integration/
         elif ("s3:path/" in uri or "s3:action/" in uri) and method == "GET":
@@ -768,14 +673,14 @@ def invoke_rest_api_integration_backend(invocation_context: ApiInvocationContext
                 account_id, queue = uri.split("/")[-2:]
                 region_name = uri.split(":")[3]
                 if "GetQueueUrl" in template or "CreateQueue" in template:
-                    new_request = (
-                        f"{aws_stack.render_velocity_template(template, data)}&QueueName={queue}"
-                    )
+                    request_templates = RequestTemplates()
+                    payload = request_templates.render(invocation_context)
+                    new_request = f"{payload}&QueueName={queue}"
                 else:
+                    request_templates = RequestTemplates()
+                    payload = request_templates.render(invocation_context)
                     queue_url = f"{config.get_edge_url()}/{account_id}/{queue}"
-                    new_request = (
-                        f"{aws_stack.render_velocity_template(template, data)}&QueueUrl={queue_url}"
-                    )
+                    new_request = f"{payload}&QueueUrl={queue_url}"
                 headers = aws_stack.mock_aws_request_headers(service="sqs", region_name=region_name)
 
                 url = urljoin(config.service_url("sqs"), f"{TEST_AWS_ACCOUNT_ID}/{queue}")
@@ -784,7 +689,10 @@ def invoke_rest_api_integration_backend(invocation_context: ApiInvocationContext
                 )
                 return result
             elif uri.startswith("arn:aws:apigateway:") and ":sns:path" in uri:
-                integration_response = SnsIntegration(invocation_context).invoke()
+                invocation_context.context = helpers.get_event_request_context(invocation_context)
+                invocation_context.stage_variables = helpers.get_stage_variables(invocation_context)
+
+                integration_response = SnsIntegration().invoke(invocation_context)
                 return apply_request_response_templates(
                     integration_response, response_templates, content_type=APPLICATION_JSON
                 )
@@ -842,21 +750,33 @@ def invoke_rest_api_integration_backend(invocation_context: ApiInvocationContext
                 uri = "http://%s/%s" % (instance["Id"], invocation_path.lstrip("/"))
 
         # apply custom request template
-        data = apply_template(integration, "request", data)
-        if isinstance(data, dict):
-            data = json.dumps(data)
+        invocation_context.context = helpers.get_event_request_context(invocation_context)
+        invocation_context.stage_variables = helpers.get_stage_variables(invocation_context)
+        request_templates = RequestTemplates()
+        payload = request_templates.render(invocation_context)
+
+        if isinstance(payload, dict):
+            payload = json.dumps(payload)
         uri = apply_request_parameters(
             uri, integration=integration, path_params=path_params, query_params=query_string_params
         )
-        result = requests.request(method=method, url=uri, data=data, headers=headers)
+        result = requests.request(method=method, url=uri, data=payload, headers=headers)
         # apply custom response template
-        result = apply_template(integration, "response", result)
-        return result
+        invocation_context.response = result
+        response_templates = ResponseTemplates()
+        response_templates.render(invocation_context)
+        return invocation_context.response
 
     elif integration_type == "MOCK":
-        # return empty response - details filled in via responseParameters above...
+
+        # TODO: apply tell don't ask principle inside ResponseTemplates or InvocationContext
+        invocation_context.stage_variables = helpers.get_stage_variables(invocation_context)
         invocation_context.response = requests_response({})
-        return apply_integration_response_template(invocation_context)
+
+        response_templates = ResponseTemplates()
+        response_templates.render(invocation_context)
+
+        return invocation_context.response
 
     if method == "OPTIONS":
         # fall back to returning CORS headers if this is an OPTIONS request
@@ -868,97 +788,6 @@ def invoke_rest_api_integration_backend(invocation_context: ApiInvocationContext
     )
 
 
-def get_target_resource_details(invocation_context: ApiInvocationContext) -> Tuple[str, Dict]:
-    """Look up and return the API GW resource (path pattern + resource dict) for the given invocation context."""
-    path_map = helpers.get_rest_api_paths(
-        rest_api_id=invocation_context.api_id, region_name=invocation_context.region_name
-    )
-    relative_path = invocation_context.invocation_path
-    try:
-        extracted_path, resource = get_resource_for_path(path=relative_path, path_map=path_map)
-        invocation_context.resource = resource
-        return extracted_path, resource
-    except Exception:
-        return None, None
-
-
-def get_target_resource_method(invocation_context: ApiInvocationContext) -> Optional[Dict]:
-    """Look up and return the API GW resource method for the given invocation context."""
-    _, resource = get_target_resource_details(invocation_context)
-    if not resource:
-        return None
-    methods = resource.get("resourceMethods") or {}
-    method_name = invocation_context.method.upper()
-    method_details = methods.get(method_name) or methods.get("ANY")
-    return method_details
-
-
-def get_stage_variables(api_id: str, stage: str) -> Dict[str, str]:
-    if not stage:
-        return {}
-    region_name = [name for name, region in apigateway_backends.items() if api_id in region.apis][0]
-    api_gateway_client = aws_stack.connect_to_service("apigateway", region_name=region_name)
-    try:
-        response = api_gateway_client.get_stage(restApiId=api_id, stageName=stage)
-        return response.get("variables")
-    except Exception:
-        LOG.info(f"Failed to get stage {stage} for api id {api_id}")
-        return {}
-
-
-def get_event_request_context(invocation_context: ApiInvocationContext):
-    method = invocation_context.method
-    path = invocation_context.path
-    headers = invocation_context.headers
-    integration_uri = invocation_context.integration_uri
-    resource_path = invocation_context.resource_path
-    resource_id = invocation_context.resource_id
-
-    set_api_id_stage_invocation_path(invocation_context)
-    relative_path, query_string_params = extract_query_string_params(
-        path=invocation_context.path_with_query_string
-    )
-    api_id = invocation_context.api_id
-    stage = invocation_context.stage
-
-    source_ip = headers.get("X-Forwarded-For", ",").split(",")[-2].strip()
-    integration_uri = integration_uri or ""
-    account_id = integration_uri.split(":lambda:path")[-1].split(":function:")[0].split(":")[-1]
-    account_id = account_id or TEST_AWS_ACCOUNT_ID
-    domain_name = f"{api_id}.execute-api.{LOCALHOST_HOSTNAME}"
-    request_context = {
-        "resourcePath": resource_path or relative_path,
-        "apiId": api_id,
-        "domainPrefix": api_id,
-        "domainName": domain_name,
-        "accountId": account_id,
-        "resourceId": resource_id,
-        "requestId": long_uid(),
-        "identity": {
-            "accountId": account_id,
-            "sourceIp": source_ip,
-            "userAgent": headers.get("User-Agent"),
-        },
-        "httpMethod": method,
-        "protocol": "HTTP/1.1",
-        "requestTime": pytz.utc.localize(datetime.datetime.utcnow()).strftime(
-            REQUEST_TIME_DATE_FORMAT
-        ),
-        "requestTimeEpoch": int(time.time() * 1000),
-    }
-
-    # set "authorizer" and "identity" event attributes from request context
-    auth_context = invocation_context.auth_context
-    if auth_context:
-        request_context["authorizer"] = auth_context
-    request_context["identity"].update(invocation_context.auth_identity or {})
-
-    if not is_test_invoke_method(method, path):
-        request_context["path"] = (f"/{stage}" if stage else "") + relative_path
-        request_context["stage"] = stage
-    return request_context
-
-
 def apply_request_response_templates(
     data: Union[Response, bytes],
     templates: Dict[str, str],
@@ -974,7 +803,7 @@ def apply_request_response_templates(
     if not template:
         return data
     content = (data.content if is_response else data) or ""
-    result = aws_stack.render_velocity_template(template, content, as_json=as_json)
+    result = VtlTemplate().render_vtl(template, content, as_json=as_json)
     if is_response:
         data._content = result
         update_content_length(data)
@@ -982,9 +811,5 @@ def apply_request_response_templates(
     return result
 
 
-def is_test_invoke_method(method, path):
-    return method == "POST" and bool(re.match(PATH_REGEX_TEST_INVOKE_API, path))
-
-
 # instantiate listener
 UPDATE_APIGATEWAY = ProxyListenerApiGateway()
